#include "iconhadler.h"
#include <QtCore/qendian.h>
#include <QtGui/QImage>
#include <QtCore/QFile>
#include <QtCore/QBuffer>


bool writeIconDirEntry(QIODevice *iodev, const ICONDIRENTRY &iconEntry)
{
    if (iodev) {
        uchar tmp[ICONDIRENTRY_SIZE];
        tmp[0] = iconEntry.bWidth;
        tmp[1] = iconEntry.bHeight;
        tmp[2] = iconEntry.bColorCount;
        tmp[3] = iconEntry.bReserved;
        qToLittleEndian<quint16>(iconEntry.wPlanes, &tmp[4]);
        qToLittleEndian<quint16>(iconEntry.wBitCount, &tmp[6]);
        qToLittleEndian<quint32>(iconEntry.dwBytesInRes, &tmp[8]);
        qToLittleEndian<quint32>(iconEntry.dwImageOffset, &tmp[12]);
        return (iodev->write((char*)tmp,  ICONDIRENTRY_SIZE) == ICONDIRENTRY_SIZE) ? true : false;
    }

    return false;
}

bool writeIconDir(QIODevice *iodev, const ICONDIR &iconDir)
{
    if (iodev) {
        uchar tmp[6];
        qToLittleEndian(iconDir.idReserved, tmp);
        qToLittleEndian(iconDir.idType, &tmp[2]);
        qToLittleEndian(iconDir.idCount, &tmp[4]);
        return (iodev->write((char*)tmp,  6) == 6) ? true : false;
    }
    return false;
}

bool writeBMPInfoHeader(QIODevice *iodev, const BMP_INFOHDR &header)
{
    if (iodev) {
        uchar tmp[BMP_INFOHDR_SIZE];
        qToLittleEndian<quint32>(header.biSize, &tmp[0]);
        qToLittleEndian<quint32>(header.biWidth, &tmp[4]);
        qToLittleEndian<quint32>(header.biHeight, &tmp[8]);
        qToLittleEndian<quint16>(header.biPlanes, &tmp[12]);
        qToLittleEndian<quint16>(header.biBitCount, &tmp[14]);
        qToLittleEndian<quint32>(header.biCompression, &tmp[16]);
        qToLittleEndian<quint32>(header.biSizeImage, &tmp[20]);
        qToLittleEndian<quint32>(header.biXPelsPerMeter, &tmp[24]);
        qToLittleEndian<quint32>(header.biYPelsPerMeter, &tmp[28]);
        qToLittleEndian<quint32>(header.biClrUsed, &tmp[32]);
        qToLittleEndian<quint32>(header.biClrImportant, &tmp[36]);

        return (iodev->write((char*)tmp, BMP_INFOHDR_SIZE) == BMP_INFOHDR_SIZE) ? true : false;
    }
    return false;
}



bool writeIco(QIODevice * device, const QList<QImage> & images)
{
    bool retValue = false;

    if (images.count()) {

        qint64 origOffset = device->pos();

        ICONDIR id;
        id.idReserved = 0;
        id.idType = 1;
        id.idCount = images.count();

        ICONDIRENTRY * entries = new ICONDIRENTRY[id.idCount];
        BMP_INFOHDR * bmpHeaders = new BMP_INFOHDR[id.idCount];
        QByteArray * imageData = new QByteArray[id.idCount];

        for (int i=0; i<id.idCount; i++) {

            QImage image = images[i];
            // Scale down the image if it is larger than 128 pixels in either width or height
            if (image.width() > 128 || image.height() > 128)
            {
                image = image.scaled(128, 128, Qt::KeepAspectRatio, Qt::SmoothTransformation);
            }
            QImage maskImage(image.width(), image.height(), QImage::Format_Mono);
            image = image.convertToFormat(QImage::Format_ARGB32);

            if (image.hasAlphaChannel()) {
                maskImage = image.createAlphaMask();
            } else {
                maskImage.fill(0xff);
            }
            maskImage = maskImage.convertToFormat(QImage::Format_Mono);

            int	nbits = 32;
            int	bpl_bmp = ((image.width()*nbits+31)/32)*4;

            entries[i].bColorCount = 0;
            entries[i].bReserved = 0;
            entries[i].wBitCount = nbits;
            entries[i].bHeight = image.height();
            entries[i].bWidth = image.width();
            entries[i].dwBytesInRes = BMP_INFOHDR_SIZE + (bpl_bmp * image.height())
                + (maskImage.bytesPerLine() * maskImage.height());
            entries[i].wPlanes = 1;
            if (i==0)
                entries[i].dwImageOffset = origOffset + ICONDIR_SIZE
                + (id.idCount * ICONDIRENTRY_SIZE);
            else
                entries[i].dwImageOffset = entries[i-1].dwImageOffset + entries[i-1].dwBytesInRes;

            bmpHeaders[i].biBitCount = entries[i].wBitCount;
            bmpHeaders[i].biClrImportant = 0;
            bmpHeaders[i].biClrUsed = entries[i].bColorCount;
            bmpHeaders[i].biCompression = 0;
            bmpHeaders[i].biHeight = entries[i].bHeight * 2; // 2 is for the mask
            bmpHeaders[i].biPlanes = entries[i].wPlanes;
            bmpHeaders[i].biSize = BMP_INFOHDR_SIZE;
            bmpHeaders[i].biSizeImage = entries[i].dwBytesInRes - BMP_INFOHDR_SIZE;
            bmpHeaders[i].biWidth = entries[i].bWidth;
            bmpHeaders[i].biXPelsPerMeter = 0;
            bmpHeaders[i].biYPelsPerMeter = 0;

            QBuffer buffer(&imageData[i]);
            buffer.open(QIODevice::WriteOnly);

            uchar *buf = new uchar[bpl_bmp];
            uchar *b;
            memset( buf, 0, bpl_bmp );
            int y;
            for (y=image.height()-1; y>=0; y--) {	// write the image bits
                // 32 bits
                QRgb *p   = (QRgb *)image.scanLine(y);
                QRgb *end = p + image.width();
                b = buf;
                int x = 0;
                while (p < end) {
                    *b++ = qBlue(*p);
                    *b++ = qGreen(*p);
                    *b++ = qRed(*p);
                    *b++ = qAlpha(*p);
                    if (qAlpha(*p) > 0)   // Even mostly transparent pixels must not be masked away
                        maskImage.setPixel(x, y, Qt::color1);  // (i.e. createAlphaMask() takes away too much)
                    p++;
                    x++;
                }
                buffer.write((char*)buf, bpl_bmp);
            }
            delete[] buf;

            maskImage.invertPixels();   // seems as though it needs this
            // NOTE! !! The mask is only flipped vertically - not horizontally !!
            for (y=maskImage.height()-1; y>=0; y--)
                buffer.write((char*)maskImage.scanLine(y), maskImage.bytesPerLine());

        }



        if (writeIconDir(device, id)) {
            int i;
            bool bOK = true;
            for (i = 0; i < id.idCount && bOK; i++) {
                bOK = writeIconDirEntry(device, entries[i]);
            }
            if (bOK) {
                for (i=0; i<id.idCount && bOK; i++) {
                    bOK = writeBMPInfoHeader(device, bmpHeaders[i]);
                    bOK &= (device->write(imageData[i]) == (int) imageData[i].size());
                }
                retValue = bOK;
            }
        }

        delete [] entries;
        delete [] bmpHeaders;
        delete [] imageData;

    }
    return retValue;
}
